@@ -1,9 +1,9 @@
 /*
  * This file is part of the Nice GLib ICE library.
  *
- * (C) 2006, 2007 Collabora Ltd.
- *  Contact: Dafydd Harries
- * (C) 2006, 2007 Nokia Corporation. All rights reserved.
+ * (C) 2006-2010 Collabora Ltd.
+ *  Contact: Youness Alaoui
+ * (C) 2006-2010 Nokia Corporation. All rights reserved.
  *  Contact: Kai Vehmanen
  *
  * The contents of this file are subject to the Mozilla Public License Version
@@ -23,6 +23,7 @@
  *
  * Contributors:
  *   Dafydd Harries, Collabora Ltd.
+ *   Youness Alaoui, Collabora Ltd.
  *   Kai Vehmanen, Nokia
  *
  * Alternatively, the contents of this file may be used under the terms of the
@@ -236,6 +237,8 @@ typedef enum
  * An enum to specify which proxy type to use for relaying.
  * Note that the proxies will only be used with TCP TURN relaying.
  * <para> See also: #NiceAgent:proxy-type </para>
+ *
+ * Since: 0.0.4
  */
 typedef enum
 {
@@ -277,14 +280,38 @@ typedef void (*NiceAgentRecvFunc) (
 NiceAgent *
 nice_agent_new (GMainContext *ctx, NiceCompatibility compat);
 
+
+/**
+ * nice_agent_new_reliable:
+ * @ctx: The Glib Mainloop Context to use for timers
+ * @compat: The compatibility mode of the agent
+ *
+ * Create a new #NiceAgent in reliable mode, which uses #PseudoTcpSocket to
+ * assure reliability of the messages.
+ * The returned object must be freed with g_object_unref()
+ * <para> See also: #NiceAgent::reliable-transport-writable </para>
+ *
+ * Since: 0.0.11
+ *
+ * Returns: The new agent GObject
+ */
+NiceAgent *
+nice_agent_new_reliable (GMainContext *ctx, NiceCompatibility compat);
+
 /**
  * nice_agent_add_local_address:
  * @agent: The #NiceAgent Object
  * @addr: The address to listen to
  * If the port is 0, then a random port will be chosen by the system
  *
- * Add a local address from which to derive local host candidates
+ * Add a local address from which to derive local host candidates for
+ * candidate gathering.
+ * <para>
+ * Since 0.0.5, if this method is not called, libnice will automatically
+ * discover the local addresses available
+ * </para>
  *
+ * See also: nice_agent_gather_candidates()
  * Returns: %TRUE on success, %FALSE on fatal (memory allocation) errors
  */
 gboolean
@@ -352,13 +379,15 @@ gboolean nice_agent_set_relay_info(
  * Start the candidate gathering process.
  * Once done, #NiceAgent::candidate-gathering-done is called for the stream
  *
+ * See also: nice_agent_add_local_address()
  <note>
    <para>
     Local addresses can be previously set with nice_agent_add_local_address()
   </para>
   <para>
-    If no local address was previously added, then the nice agent will
-    automatically detect the local address using nice_interfaces_get_local_ips()
+    Since 0.0.5, If no local address was previously added, then the nice agent
+    will automatically detect the local address using
+    nice_interfaces_get_local_ips()
    </para>
  </note>
  */
@@ -488,7 +517,7 @@ nice_agent_send (
      The caller owns the returned GSList as well as the candidates contained
      within it.
      To get full results, the client should wait for the
-     #NiceAgent::candidates-gathering-done signal.
+     #NiceAgent::candidate-gathering-done signal.
    </para>
  </note>
  *
@@ -629,6 +658,7 @@ nice_agent_set_selected_remote_candidate (
  *
  * Sets the IP_TOS and/or IPV6_TCLASS field on the stream's sockets' options
  *
+ * Since: 0.0.9
  */
 void nice_agent_set_stream_tos (
   NiceAgent *agent,
@@ -647,6 +677,7 @@ void nice_agent_set_stream_tos (
  * <para>
  * The SOFTWARE attribute will only be added in the #NICE_COMPATIBILITY_DRAFT19
  * and #NICE_COMPATIBILITY_WLM2009 compatibility modes.
+ *
  * </para>
  * <note>
      <para>
@@ -659,6 +690,8 @@ void nice_agent_set_stream_tos (
      </para>
    </note>
  *
+ * Since: 0.0.10
+ *
  */
 void nice_agent_set_software (NiceAgent *agent, const gchar *software);
 
